import datetime as dt
import io
import math
import mimetypes
import string
import time
from collections import namedtuple
from decimal import Decimal
from itertools import groupby
from statistics import mean

import pytz
from pdf2image import convert_from_bytes
from PIL import Image
from sql import Literal
from sql.aggregate import Count

from trytond.config import config
from trytond.i18n import gettext
from trytond.model import (
    Index, ModelSQL, ModelView, Unique, ValueMixin, Workflow, fields)
from trytond.modules.company.model import employee_field
from trytond.modules.product import price_digits, round_price
from trytond.modules.sale_shipment_cost.stock import ShipmentCostSaleMixin
from trytond.modules.stock_package.exceptions import PackageError
from trytond.modules.stock_package.stock import PackageMixin
from trytond.modules.stock_package_shipping.stock import ShippingMixin
from trytond.modules.stock_package_shipping_dpd.stock import ShippingDPDMixin
from trytond.modules.stock_package_shipping_sendcloud.stock import (
    ShippingSendcloudMixin)
from trytond.modules.stock_package_shipping_ups.stock import ShippingUPSMixin
from trytond.modules.stock_shipment_cost_weight.stock import (
    ShipmentCostMixinWeight)
from trytond.modules.stock_shipment_measurements.stock import MeasurementsMixin
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Bool, Eval, If
from trytond.report import Report
from trytond.tools import qrcode
from trytond.transaction import Transaction
from trytond.wizard import (
    Button, StateAction, StateTransition, StateView, Wizard)

from .mixin import NotificationMixin
from .tools import int2base, remove_emoji

webshop_base_url = config.get('webshop', 'url')


class _ShipmentEmailReport(Report):

    @classmethod
    def get_context(cls, records, header, data):
        context = super().get_context(records, header, data)
        context['group_shipments'] = cls.group_shipments
        return context

    @classmethod
    def group_shipments(cls, shipments):
        def key(shipment):
            return shipment.state == 'done'
        return groupby(sorted(shipments, key=key), key)


class Configuration(metaclass=PoolMeta):
    __name__ = 'stock.configuration'

    max_x = fields.MultiValue(fields.Integer(
            "Max X",
            domain=[
                ('max_x', '>=', 0),
                ]))
    max_y = fields.MultiValue(fields.Integer(
            "Max Y",
            domain=[
                ('max_y', '>=', 0),
                ]))

    pack_offset = fields.TimeDelta(
        "Pack Offset",
        help="The duration before which the shipments can be packed.")

    @classmethod
    def multivalue_model(cls, field):
        pool = Pool()
        if field in {'max_x', 'max_y'}:
            return pool.get('stock.configuration.position')
        return super().multivalue_model(field)


class ConfigurationPosition(ModelSQL, ValueMixin):
    "Stock Configuration Position"
    __name__ = 'stock.configuration.position'

    max_x = fields.Integer(
        "Max X",
        domain=['OR',
            ('max_x', '=', None),
            ('max_x', '>=', 0),
            ])
    max_y = fields.Integer(
        "Max Y",
        domain=['OR',
            ('max_y', '=', None),
            ('max_y', '>=', 0),
            ])


class Location(metaclass=PoolMeta):
    __name__ = 'stock.location'

    sub_locations = fields.One2Many(
        'stock.location.sub', 'location', "Sub-Locations")


class SubLocation(ModelSQL, ModelView):
    "Sub-Location"
    __name__ = 'stock.location.sub'

    name = fields.Char("Name", required=True)
    location = fields.Many2One('stock.location', "Location", required=True)


class ProductLocation(metaclass=PoolMeta):
    __name__ = 'stock.product.location'

    sub_location = fields.Many2One(
        'stock.location.sub', "Sub-Location",
        domain=[
            ('location', '=', Eval('location', -1)),
            ],
        depends=['location'])


class Move(metaclass=PoolMeta):
    __name__ = 'stock.move'

    from_sub_location = fields.Function(fields.Many2One(
            'stock.location.sub', "From Sub-Location",
            states={
                'invisible': ~Eval('from_sub_location', 0),
                }),
        'get_sub_location')
    to_sub_location = fields.Function(fields.Many2One(
            'stock.location.sub', "To Sub-Location",
            states={
                'invisible': ~Eval('to_sub_location', 0),
                }),
        'get_sub_location')
    packaging_cost_price = fields.Numeric(
        "Packaging Cost Price",
        digits=price_digits,
        states={
            'readonly': True,
            'invisible': ~Eval('packaging_cost_price'),
            })
    work_cost_price = fields.Numeric(
        "Work Cost Price",
        digits=price_digits,
        states={
            'readonly': True,
            'invisible': ~Eval('work_cost_price'),
            })

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls._allow_modify_closed_period.update([
                'packaging_cost_price',
                'work_cost_price',
                ])

    @classmethod
    def __register__(cls, module):
        table = cls.__table__()
        table_h = cls.__table_handler__(module)
        cursor = Transaction().connection.cursor()

        super().__register__(module)

        if table_h.column_exist('package_type_cost_price'):
            cursor.execute(*table.update(
                    [table.packaging_cost_price],
                    [table.package_type_cost_price]))
            table_h.drop_column('package_type_cost_price')

    def get_sub_location(self, name):
        field = {
            'from_sub_location': 'from_location',
            'to_sub_location': 'to_location',
            }[name]
        location = None
        for product_location in self.product.locations:
            if getattr(self, field) == product_location.location:
                location = product_location.sub_location
                break
        return location.id if location else None

    @property
    def origin_name(self):
        pool = Pool()
        SaleLine = pool.get('sale.line')
        Lang = pool.get('ir.lang')
        if (isinstance(self.shipment, ShipmentOut)
                and self.shipment.customer.lang):
            language = self.shipment.customer.lang.code
        else:
            language = None
        lang = Lang.get(language)
        name = super().origin_name
        if isinstance(self.origin, SaleLine):
            name = gettext(
                'jurassic.msg_account_invoice_sale_origin',
                language,
                sale=self.origin.sale.rec_name,
                date=lang.strftime(self.origin.sale.sale_date))
        return name

    def pack_key(self):
        key = super(Move, self).pack_key()
        if self.shipment:
            warehouse = self.shipment.warehouse
            for i, (name, value) in enumerate(key):
                if name == 'name':
                    key[i] = (
                        name, self.product.get_rec_name_position(
                            warehouse, excluded=False))
                    break
        key.append(
            ('classes', ('bio',) if self.product.organic else ())
            )
        return key

    @classmethod
    @ModelView.button
    @Workflow.transition('done')
    def do(cls, moves):
        super().do(moves)

        for move in moves:
            warehouse = (
                move.from_location.warehouse
                or move.to_location.warehouse)
            if not warehouse:
                continue
            package = move.package
            while package:
                for cost in move.product.package_type_costs_used(
                        package_type=package.type.id):
                    move.packaging_cost_price = cost.packaging_price
                    move.work_cost_price = cost.work_price
                    break
                else:
                    package = package.parent
                    continue
                break
        cls.save(moves)


class ShipmentDeliveryDateMixin:
    __slots__ = ()

    delivery_date = fields.Function(
        fields.Date("Delivery Date"), 'get_delivery_date')

    def get_delivery_date(self, name):
        date = self.effective_date or self.planned_date
        if self.carrier:
            return self.carrier.get_delivery_date(
                self.delivery_address, date=date)
        else:
            return date


class ShipmentOut(
        ShipmentDeliveryDateMixin, NotificationMixin, metaclass=PoolMeta):
    __name__ = 'stock.shipment.out'

    first = fields.Boolean("First", readonly=True)
    like_ripe_fruits = fields.Function(fields.Boolean(
            "Like Ripe Fruits"), 'on_change_with_like_ripe_fruits')
    shipping_hour = fields.Time("Shipping Hour", readonly=True)
    has_producible = fields.Integer("Has Producible", readonly=True)
    pack_distance = fields.Function(
        fields.Float("Distance"), 'get_pack_distance')
    notes = fields.One2Many(
        'ir.note', 'resource', "Notes", readonly=True)
    closed_by = employee_field("Closed By")
    delivery_note_printed = fields.Boolean(
        "Delivery Note Printed")
    partial = fields.Function(fields.Boolean("Partial"),
        'get_partial')
    sale_shipments = fields.Function(fields.Many2Many(
            'stock.shipment.out', None, None, "Shipments"),
        'get_sale_shipments')
    sale_drop_shipments = fields.Function(fields.Many2Many(
            'stock.shipment.drop', None, None, "Drop Shipments"),
        'get_sale_drop_shipments')
    sale_moves_cancelled = fields.Function(fields.Many2Many(
            'stock.move', None, None, "Moves Cancelled"),
        'get_sale_moves_cancelled')
    comment = fields.Char("Comment")

    @classmethod
    def default_first(cls):
        return False

    @fields.depends('customer')
    def on_change_with_like_ripe_fruits(self, name=None):
        if self.customer:
            return self.customer.like_ripe_fruits

    @classmethod
    def default_has_producible(cls):
        return 0

    @classmethod
    def default_delivery_note_printed(cls):
        return False

    @property
    def url_webshop(self):
        url = webshop_base_url
        webshops = {
            m.sale.web_shop for m in self.outgoing_moves
            if m.sale and m.sale.web_shop}
        if len(webshops) == 1:
            webshop, = webshops
            url = webshop.url
        return url

    @property
    def url_webshop_sales(self):
        sales = {m.sale for m in self.outgoing_moves if m.sale}
        if len(sales) == 1:
            sale = sales.pop()
            url = self.url_webshop + 'sale/%s' % sale.id
        else:
            url = (
                self.url_webshop + 'sales?'
                + '&'.join('id=%s' % s.id for s in sales))
        return url

    def get_pack_distance(self, name):
        pool = Pool()
        User = pool.get('res.user')
        transaction = Transaction()
        context = transaction.context

        user_id = transaction.user
        if not user_id:
            user_id = context.get('user', user_id)
        user = User(user_id)
        if user.stock_position_x is None or user.stock_position_y is None:
            return 0

        distances = []
        for move in self.outgoing_moves:
            position = move.product.get_position(self.warehouse)
            if position and position.x is not None and position.y is not None:
                distance = math.sqrt(
                    (position.x - user.stock_position_x) ** 2
                    + (position.y - user.stock_position_y) ** 2)
                distances.append(distance)
        if distances:
            return mean(distances)
        else:
            return 0

    def get_partial(self, name):
        return (len(self.sale_shipments) + len(self.sale_drop_shipments)) > 0

    def get_sale_shipments(self, name):
        sales = {m.sale for m in self.outgoing_moves if m.sale}
        shipments = {
            sh for s in sales for sh in s.shipments if sh.state != 'cancelled'}
        shipments.discard(self)
        return list(map(int, shipments))

    def get_sale_drop_shipments(self, name):
        sales = {m.sale for m in self.outgoing_moves if m.sale}
        drop_shipments = {
            d for s in sales for d in s.drop_shipments
            if d.state != 'cancelled'}
        return list(map(int, drop_shipments))

    def get_sale_moves_cancelled(self, name):
        sales = {m.sale for m in self.outgoing_moves if m.sale}
        moves_recreated = {
            m for s in sales for l in s.lines for m in l.moves_recreated}
        return [
            m.id for s in sales for m in s.moves
            if m.state == 'cancelled' and m not in moves_recreated]

    @property
    def best_pack_user(self):
        pool = Pool()
        User = pool.get('res.user')
        users = User.search([
                ('stock_position_x', '!=', None),
                ('stock_position_y', '!=', None),
                ])
        distances = []
        for user in users:
            with Transaction().set_user(user.id):
                shipment = self.__class__(self)
                distances.append((shipment.pack_distance, user))
        _, user = min(distances)
        return user

    def get_rec_name(self, name):
        name = super().get_rec_name(name)
        if self.shipping_hour:
            name += self.shipping_hour.strftime(' (%H:%M)')
        if self.first:
            name += ' [!!new!!]'
        if self.like_ripe_fruits:
            name += ' [ripe]'
        if self.delivery_address.country:
            name += ' [%s]' % self.delivery_address.country.code
        return name

    @property
    def additional_weight(self):
        weight = 0
        for package in self.packages:
            weight += package.total_weight
        return weight

    @classmethod
    def pack_search(cls, domain, limit=None, order=None):
        if order is None:
            order = []
        else:
            order = order.copy()
        order.insert(0, ('shipping_hour', 'ASC NULLS LAST'))
        order.insert(1, ('has_producible', 'ASC NULLS FIRST'))
        shipments = super().pack_search(domain, limit=limit, order=order)

        def key(shipment):
            shipping_hour = shipment.shipping_hour
            if shipping_hour is None:
                shipping_hour = dt.time.max
            return (
                shipping_hour,
                bool(shipment.has_producible),
                shipment.pack_distance)
        shipments.sort(key=key)
        return shipments

    def _pack_moves(self, key, packed):
        pool = Pool()
        Move = pool.get('stock.move')
        # Update biggest move with the extra quantity
        quantity = key.get('quantity', 0)
        available_moves = [m for m in self.outgoing_moves
            if m not in packed and m.pack_match(key)]
        available_moves.sort(key=lambda m: m.quantity)
        available_qty = sum(m.quantity for m in available_moves)

        if quantity > available_qty and available_moves:
            move = available_moves[0]
            move.quantity = move.unit.ceil(
                move.quantity + quantity - available_qty)
            move.save()
            Move.check_over_shipment([move])

        return super()._pack_moves(key, packed)

    def ship_GET(self):
        shipment = super().ship_GET()
        if self.customer.lang:
            shipment['customer'] += ' (%s)' % self.customer.lang.name
        return shipment

    def pack_GET(self):
        shipment = super().pack_GET()
        if self.customer.lang:
            shipment['customer'] += ' (%s)' % self.customer.lang.name
        return shipment

    def pack_PUT(self, data):
        if data.get('state') == 'packed':
            resources = self.resources()
            if resources['note_unread']:
                raise PackageError(
                    gettext('jurassic.msg_stock_package_note_unread',
                        shipment=self.rec_name))
        super().pack_PUT(data)

    @classmethod
    def pack_date(cls):
        pool = Pool()
        Company = pool.get('company.company')
        Config = pool.get('stock.configuration')
        super().pack_date()
        config = Config(1)
        company_id = Transaction().context.get('company')
        timezone = None
        if company_id:
            company = Company(company_id)
            if company.timezone and pytz:
                timezone = pytz.timezone(company.timezone)
        now = dt.datetime.now(timezone)
        if config.pack_offset:
            now += config.pack_offset
        return now.date()

    @classmethod
    def check_packages(cls, shipments):
        super().check_packages(shipments)
        for shipment in shipments:
            if not shipment.packages:
                raise PackageError(
                    gettext('stock_package.msg_package_mismatch',
                        shipment=shipment.rec_name))

    @classmethod
    def wait(cls, shipments, moves=None):
        if Transaction().context.get('_jf_shipment_out_skip_wait', False):
            return
        super().wait(shipments, moves=moves)

    @classmethod
    def create(cls, vlist):
        shipments = super().create(vlist)
        cls._set_has_producible(shipments)
        cls._set_shipping_hour(shipments)
        cls.save(shipments)
        return shipments

    @classmethod
    def write(cls, *args):
        shipments = sum(args[0:None:2], [])
        super().write(*args)
        cls._set_has_producible(shipments)
        cls._set_shipping_hour(shipments)
        cls._notify_sale(shipments)
        cls.save(shipments)

    @classmethod
    def _set_has_producible(cls, shipments):
        for shipment in shipments:
            count = 0
            for move in shipment.outgoing_moves:
                if move.product.producible:
                    count += 1
            if shipment.has_producible != count:
                shipment.has_producible = count

    @classmethod
    def _set_shipping_hour(cls, shipments):
        for shipment in shipments:
            shipping_hour = None
            if shipment.carrier:
                pattern = {}
                date = shipment.effective_date or shipment.planned_date
                if date:
                    pattern['day'] = date.weekday()
                shipping_hour = shipment.carrier.get_shipping_hour(
                    shipment.delivery_address, pattern=pattern)
            if shipment.shipping_hour != shipping_hour:
                shipment.shipping_hour = shipping_hour

    @classmethod
    def _notify_sale(cls, shipments):
        pool = Pool()
        Sale = pool.get('sale.sale')
        notify_delivery_date = set()
        notify_carrier = set()
        for shipment in shipments:
            for sale in {m.sale for m in shipment.outgoing_moves if m.sale}:
                if sale.delivery_date != shipment.delivery_date:
                    notify_delivery_date.add(sale)
                if sale.carrier != shipment.carrier:
                    notify_carrier.add(sale)
        Sale.write(
            list(notify_delivery_date), {
                'notify_delivery_date': True,
                },
            list(notify_carrier), {
                'notify_carrier': True,
                })


class ShipmentOutPickingList(metaclass=PoolMeta):
    __name__ = 'stock.shipment.out.picking_list'

    @classmethod
    def moves(cls, shipment):
        moves = super().moves(shipment)

        def position(move):
            position = move.product.get_position(
                shipment.warehouse, excluded=True)
            if position and position.rec_name:
                return (0, position.rec_name)
            else:
                return (1, '')
        return sorted(moves, key=position)


class ShipmentOutDeliveryNote(metaclass=PoolMeta):
    __name__ = 'stock.shipment.out.delivery_note'

    @classmethod
    def get_context(cls, records, header, data):
        context = super().get_context(records, header, data)
        context['qrcode'] = qrcode.generate_png
        context['remove_emoji'] = remove_emoji
        return context

    @classmethod
    def execute(cls, ids, data):
        pool = Pool()
        ShipmentOut = pool.get('stock.shipment.out')
        result = super().execute(ids, data)
        with Transaction().new_transaction():
            ShipmentOut.write(ShipmentOut.browse(ids), {
                    'delivery_note_printed': True,
                    })
        return result


class ShipmentOutDone(_ShipmentEmailReport):
    __name__ = 'stock.shipment.out.done'


class ShipmentOutClosedBySet(Wizard):
    "Set Closed By"
    __name__ = 'stock.shipment.out.closed_by.set'

    start = StateView(
        'stock.shipment.out.closed_by.set.start',
        'jurassic.stock_shipment_out_closed_set_start_view_form', [
            Button("Cancel", 'end', 'tryton-cancel'),
            Button("Set", 'set', 'tryton-ok', default=True),
            ])
    set = StateTransition()

    def transition_set(self):
        pool = Pool()
        Shipment = pool.get('stock.shipment.out')
        for shipment in self.start.shipments:
            shipment.closed_by = self.start.employee
        Shipment.save(self.start.shipments)
        return 'end'


class ShipmentOutClosedBySetStart(ModelView):
    "Set Closed By"
    __name__ = 'stock.shipment.out.closed_by.set.start'

    employee = fields.Many2One(
        'company.employee', "Employee", required=True,
        domain=[('company', '=', Eval('context', {}).get('company', -1))],
        )
    shipments = fields.Many2Many(
        'stock.shipment.out', None, None, "Shipments",
        domain=[('company', '=', Eval('context', {}).get('company', -1))])


class ShipmentCostMixin:
    # Use a copy fro stock_shipment_cost to avoid set_cost method collision
    __slots__ = ()

    carrier = fields.Many2One(
        'carrier', "Carrier",
        states={
            'readonly': Eval('shipment_cost_readonly', True),
            },
        context={
            'company': Eval('company', -1),
            },
        depends=['company'])

    cost_currency_used = fields.Function(fields.Many2One(
            'currency.currency', "Cost Currency",
            states={
                'readonly': (
                    Eval('shipment_cost_readonly', True)
                    | ~Eval('cost_edit', False)),
                }),
        'on_change_with_cost_currency_used', setter='set_cost_used')
    cost_currency = fields.Many2One(
        'currency.currency', "Cost Currency",
        states={
            'required': Bool(Eval('cost')),
            'readonly': Eval('shipment_cost_readonly', True),
            })

    cost_used = fields.Function(fields.Numeric(
            "Cost", digits=price_digits,
            states={
                'readonly': (
                    Eval('shipment_cost_readonly', True)
                    | ~Eval('cost_edit', False)),
                }),
        'on_change_with_cost_used', setter='set_cost_used')
    cost = fields.Numeric("Cost", digits=price_digits, readonly=True)
    cost_edit = fields.Boolean(
        "Edit Cost",
        states={
            'readonly': Eval('shipment_cost_readonly', True),
            'invisible': Eval('shipment_cost_readonly', True),
            },
        help="Check to edit the cost.")

    shipment_cost_readonly = fields.Function(
        fields.Boolean("Shipment Cost Read Only"),
        'on_change_with_shipment_cost_readonly')

    def _get_carrier_context(self):
        return {}

    def get_carrier_context(self):
        return self._get_carrier_context()

    @fields.depends('carrier', 'company', methods=['_get_carrier_context'])
    def _compute_costs(self):
        costs = {
            'cost': None,
            'cost_currency': None,
            }
        if self.carrier:
            with Transaction().set_context(self._get_carrier_context()):
                cost, currency_id = self.carrier.get_purchase_price()
            if cost is not None:
                costs['cost'] = round_price(cost)
                costs['cost_currency'] = currency_id
        return costs

    @fields.depends(
        'cost_currency', 'cost_edit',
        methods=['_compute_costs', 'on_change_with_shipment_cost_readonly'])
    def on_change_with_cost_currency_used(self, name=None):
        readonly = self.on_change_with_shipment_cost_readonly()
        if not self.cost_edit and not readonly:
            return self._compute_costs()['cost_currency']
        elif self.cost_currency:
            return self.cost_currency

    @fields.depends(
        'cost', 'cost_edit',
        methods=['_compute_costs', 'on_change_with_shipment_cost_readonly'])
    def on_change_with_cost_used(self, name=None):
        cost = self.cost
        readonly = self.on_change_with_shipment_cost_readonly()
        if not self.cost_edit and not readonly:
            cost = self._compute_costs()['cost']
        return cost

    @classmethod
    def set_cost_used(cls, lines, name, value):
        if name.endswith('_used'):
            name = name[:-len('_used')]
        cls.write([l for l in lines if l.cost_edit], {
                name: value,
                })

    def on_change_with_shipment_cost_readonly(self, name=None):
        raise NotImplementedError

    @classmethod
    def index_set_field(cls, name):
        index = super().index_set_field(name)
        if name == 'cost_used':
            index = cls.index_set_field('cost_currency_used') + 1
        return index

    @property
    def shipment_cost_moves(self):
        raise NotImplementedError

    def _get_shipment_cost(self):
        "Return the cost for the shipment in the company currency"
        pool = Pool()
        Currency = pool.get('currency.currency')
        if self.cost_used:
            return Currency.compute(
                self.cost_currency_used, self.cost_used,
                self.company.currency, round=False)
        else:
            return Decimal(0)


class ShipmentOutCostMixin(ShipmentCostMixin):

    @fields.depends('state')
    def on_change_with_shipment_cost_readonly(self, name=None):
        return self.state in {'done', 'cancelled'}

    @classmethod
    def set_shipment_cost(cls, shipments):
        pool = Pool()
        Move = pool.get('stock.move')
        moves = []
        for shipment in shipments:
            cost = shipment._get_shipment_cost()
            if cost is None:
                continue
            factors = getattr(shipment,
                '_get_allocation_shipment_cost_factors_by_%s' %
                shipment.allocation_shipment_cost_method)()
            moves.extend(shipment._allocate_shipment_cost(cost, factors))
        Move.save(moves)

    @property
    def allocation_shipment_cost_method(self):
        if self.carrier:
            return self.carrier.shipment_cost_allocation_method
        return 'cost'

    def _get_allocation_shipment_cost_factors_by_cost(self):
        sum_ = Decimal(0)
        costs = {}
        for move in self.shipment_cost_moves:
            move_cost = (
                (move.cost_price or 0) * Decimal(str(move.internal_quantity)))
            costs[move] = move_cost
            sum_ += move_cost

        factors = {}
        for move in self.shipment_cost_moves:
            if sum_:
                ratio = costs[move] / sum_
            else:
                ratio = Decimal(1) / len(self.shipment_cost_moves)
            factors[move.id] = ratio
        return factors

    def _allocate_shipment_cost(self, cost, factors):
        moves = []
        for move in self.shipment_cost_moves:
            ratio = factors[move.id]
            quantity = Decimal(str(move.internal_quantity))
            if quantity and ratio:
                cost_price = round_price(cost * ratio / quantity)
            else:
                cost_price = Decimal(0)
            if move.shipment_out_cost_price != cost_price:
                move.shipment_out_cost_price = cost_price
                moves.append(move)
        return moves

    @classmethod
    @ModelView.button
    @Workflow.transition('done')
    def done(cls, shipments):
        for shipment in shipments:
            shipment.cost = shipment.cost_used
            shipment.cost_currency = shipment.cost_currency_used
        cls.save(shipments)
        super().done(shipments)
        cls.set_shipment_cost(shipments)


class ShipmentDrop(
        ShippingDPDMixin, ShippingUPSMixin, ShippingSendcloudMixin,
        ShippingMixin, MeasurementsMixin, PackageMixin,
        ShipmentDeliveryDateMixin, NotificationMixin,
        ShipmentCostSaleMixin, ShipmentCostMixinWeight, ShipmentOutCostMixin,
        metaclass=PoolMeta):
    __name__ = 'stock.shipment.drop'

    carrier = fields.Many2One(
        'carrier', "Carrier",
        states={
            'readonly': Eval('state') != 'draft',
            },
        depends=['state'])
    cost = fields.Numeric(
        "Cost", digits=price_digits, readonly=True)
    shipment_costs = fields.Many2Many(
        'account.shipment_cost-stock.shipment.drop',
        'shipment', 'shipment_cost', "Shipment Costs", readonly=True)
    partial = fields.Function(fields.Boolean("Partial"),
        'get_partial')
    sale_shipments = fields.Function(fields.Many2Many(
            'stock.shipment.out', None, None, "Shipments"),
        'get_sale_shipments')
    sale_drop_shipments = fields.Function(fields.Many2Many(
            'stock.shipment.drop', None, None, "Drop Shipments"),
        'get_sale_drop_shipments')
    sale_moves_cancelled = fields.Function(fields.Many2Many(
            'stock.move', None, None, "Moves Cancelled"),
        'get_sale_moves_cancelled')

    @classmethod
    def __setup__(cls):
        super().__setup__()
        package_readonly = Eval('state') != 'waiting'
        for field in [cls.packages, cls.root_packages]:
            field.states['readonly'] = package_readonly
        cls._buttons.update({
                'create_shipping': {
                    'invisible': (Eval('reference', False)
                        | ~Eval('carrier', False)),
                    'readonly': (Eval('reference', False)
                        | ~Eval('root_packages', False)
                        | ~Eval('carrier', False)
                        | ~Eval('state').in_(['waiting', 'shipped', 'done'])),
                    'depends': [
                        'reference', 'carrier', 'root_packages', 'state'],
                    },
                })

        for field in [
                cls.cost_sale_currency_used,
                cls.cost_used,
                cls.cost_sale_currency_used,
                cls.cost_sale_used,
                ]:
            field.setter = 'set_cost_used'

    @fields.depends('state')
    def on_change_with_shipment_cost_readonly(self, name=None):
        return self.state in {'done', 'cancelled'}

    @fields.depends('state')
    def on_change_with_shipment_cost_sale_readonly(self, name=None):
        return self.state in {'done', 'cancelled'}

    @property
    def shipment_cost_moves(self):
        return self.customer_moves

    @classmethod
    def set_cost_used(cls, lines, name, value):
        if name.endswith('_used'):
            name = name[:-len('_used')]
        cls.write([l for l in lines if l.cost_edit], {
                name: value,
                })

    @classmethod
    def default_cost_method(cls):
        return 'shipment'

    @property
    def _shipment_cost_currency_date(self):
        return self.effective_date

    @property
    def url_webshop(self):
        url = webshop_base_url
        webshops = {
            m.sale.web_shop for m in self.customer_moves
            if m.sale and m.sale.web_shop}
        if len(webshops) == 1:
            webshop, = webshops
            url = webshop.url
        return url

    def get_partial(self, name):
        return (len(self.sale_shipments) + len(self.sale_drop_shipments)) > 0

    def get_sale_shipments(self, name):
        sales = {m.sale for m in self.customer_moves if m.sale}
        shipments = {
            sh for s in sales for sh in s.shipments if sh.state != 'cancelled'}
        return list(map(int, shipments))

    def get_sale_drop_shipments(self, name):
        sales = {m.sale for m in self.customer_moves if m.sale}
        drop_shipments = {
            d for s in sales for d in s.drop_shipments
            if d.state != 'cancelled'}
        drop_shipments.discard(self)
        return list(map(int, drop_shipments))

    def get_sale_moves_cancelled(self, name):
        sales = {m.sale for m in self.customer_moves if m.sale}
        moves_recreated = {
            m for s in sales for l in s.lines for m in l.moves_recreated}
        return [
            m.id for s in sales for m in s.moves
            if m.state == 'cancelled' and m not in moves_recreated]

    def _get_customer_move(self, move):
        pool = Pool()
        PurchaseLine = pool.get('purchase.line')
        customer_move = super()._get_customer_move(move)
        if isinstance(move.origin, PurchaseLine):
            purchase_line = move.origin
            if len(purchase_line.requests) == 1:
                request, = purchase_line.requests
                if len(request.sale_lines) == 1:
                    sale_line, = request.sale_lines
                    if (sale_line.unit
                            and customer_move.unit
                            and (sale_line.unit.category
                                == customer_move.unit.category)):
                        customer_move.origin = sale_line
        return customer_move

    @property
    def origins(self):
        return ', '.join(set(filter(None,
                    (m.origin_name for m in self.customer_moves))))

    @classmethod
    def _measurements_location_condition(cls, shipment, move, location):
        return (
            (move.from_location == location.id)
            & (location.type == 'supplier'))

    @classmethod
    @ModelView.button_action(
        'stock_package_shipping.act_create_shipping_wizard')
    def create_shipping(cls, shipments):
        pass

    @property
    def packages_moves(self):
        return (m for m in self.customer_moves if m.state != 'cancelled')

    @property
    def shipping_allowed(self):
        return {'waiting', 'shipped', 'done'}

    @property
    def shipping_warehouse(self):
        pool = Pool()
        Location = pool.get('stock.location')
        warehouse, = Location.search([
                ('type', '=', 'warehouse'),
                ], limit=1)
        return warehouse

    @property
    def shipping_to(self):
        return self.customer

    @property
    def shipping_to_address(self):
        return self.delivery_address

    def do_create_shipping(self):
        pool = Pool()
        Session = pool.get('ir.session.wizard')

        self.__class__.create_shipping([self])

        if self.carrier and self.carrier.shipping_service:
            Wizard = pool.get(
                'stock.shipment.create_shipping.%s'
                % self.carrier.shipping_service,
                type='wizard')
            session = Session()
            session.save()
            wizard = Wizard(session.id)
            with Transaction().set_context(
                    active_id=self.id,
                    active_ids=[self.id],
                    active_model=self.__name__):
                wizard.transition_start()
            Session.delete([session])


class ShipmentDropDone(_ShipmentEmailReport):
    __name__ = 'stock.shipment.drop.done'


class ShipmentDropMakePackageShipping(Wizard):
    "Make Package and Shipping"
    __name__ = 'stock.shipment.drop.make_package_shipping'

    start = StateView(
        'stock.shipment.drop.make_package_shipping.start',
        'jurassic.stock_shipment_drop_make_package_shipping_start_view_form', [
            Button("Cancel", 'end', 'tryton-cancel'),
            Button("Make", 'make', 'tryton-ok', default=True),
            ])
    make = StateAction(
        'jurassic.stock_shipment_package_shipping_label_wizard')

    def do_make(self, action):
        pool = Pool()
        Shipment = pool.get('stock.shipment.drop')
        Package = pool.get('stock.package')
        shipment_ids = Transaction().context['active_ids']

        shipments = Shipment.browse(shipment_ids)
        packages = []
        for shipment in shipments:
            if not shipment.packages:
                package = Package()
                package.type = self.start.package_type
                package.on_change_type()
                package.shipment = shipment
                package.moves = shipment.customer_moves
                packages.append(package)
        Package.save(packages)

        for shipment in shipments:
            if not shipment.reference:
                shipment.do_create_shipping()
                time.sleep(2)

        data = {'ids': shipment_ids, 'model': Shipment.__name__}
        return action, data


class ShipmentDropMakePackageShippingStart(ModelView):
    "Make Package and Shipping"
    __name__ = 'stock.shipment.drop.make_package_shipping.start'

    package_type = fields.Many2One(
        'stock.package.type', "Package Type", required=True)

    @classmethod
    def default_package_type(cls):
        pool = Pool()
        PackageType = pool.get('stock.package.type')
        package_types = PackageType.search([])
        if len(package_types) == 1:
            package_type, = package_types
            return package_type.id


class ShipmentCostSale(metaclass=PoolMeta):
    __name__ = 'stock.shipment.cost_sale'

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls.shipment.selection.append(('stock.shipment.drop', "Drop"))


class ShipmentCreateShippingUPS(metaclass=PoolMeta):
    __name__ = 'stock.shipment.create_shipping.ups'

    def get_shipping_party(self, party, address):
        shipping_party = super().get_shipping_party(party, address)
        shipping_party['AttentionName'] = address.party_full_name[:35]
        return shipping_party


class ShipmentCreateShippingSendcloud(metaclass=PoolMeta):
    __name__ = 'stock.shipment.create_shipping.sendcloud'

    def get_parcel(self, shipment, package, credential):
        parcel = super().get_parcel(shipment, package, credential)
        parcel['company_name'] = None
        return parcel


def get_ups_code(lang, country):
    if lang == 'de':
        if country not in {'DE', 'AT'}:
            country = 'DE'
    elif lang == 'fr':
        if country not in {'FR', 'BE', 'LU', 'CH'}:
            country = 'FR'
    return '%s_%s' % (lang, country)


class Package(metaclass=PoolMeta):
    __name__ = 'stock.package'

    @classmethod
    def _get_shipment(cls):
        return super()._get_shipment() + [
            'stock.shipment.drop',
            ]

    @property
    def shipping_service(self):
        if self.shipment.carrier:
            return self.shipment.carrier.shipping_service_string

    def ship_GET_label(self):
        pool = Pool()
        LabelReport = pool.get('stock.package.shipping_label', type='report')
        super().ship_GET_label()
        oext, data, _, name = LabelReport.execute([self.id], {})
        mimetype = mimetypes.guess_type('.'.join((name, oext)))[0]
        return data, mimetype

    def shipping_label_image(
            self, orientation='portrait', width=None, height=None):
        assert orientation in {'portrait', 'landscape'}
        image = None
        if (self.shipment.carrier.shipping_service == 'ups'
                and self.shipping_label_mimetype == 'image/gif'):
            data = io.BytesIO()
            image = Image.open(io.BytesIO(self.shipping_label))
            if orientation == 'portrait':
                image = image.rotate(-90, expand=True)
            image.save(data, format='gif')
            shipping_label = data.getvalue()
        elif (self.shipment.carrier.shipping_service in {'dpd', 'sendcloud'}
                and self.shipping_label_mimetype == 'application/pdf'):
            data = io.BytesIO()
            image = convert_from_bytes(self.shipping_label, dpi=300)[0]
            if orientation == 'landscape':
                image = image.rotate(90, expand=True)
            image.save(data, format='png')
            shipping_label = data.getvalue()
        else:
            shipping_label = self.shipping_label
        if image:
            img_width, img_height = image.size
            if width:
                height = img_height * width / img_width
            elif height:
                width = img_width * height / img_height
        if width:
            width = '%scm' % width
        if height:
            height = '%scm' % height
        return (
            shipping_label, self.shipping_label_image_mimetype, width, height)

    @property
    def shipping_label_image_mimetype(self):
        if self.shipping_label_mimetype == 'application/pdf':
            return 'image/png'
        return self.shipping_label_mimetype


class PackageShippingLabel(metaclass=PoolMeta):
    __name__ = 'stock.package.shipping_label'

    @classmethod
    def render(cls, *args, **kwargs):
        return Report.render(*args, **kwargs)

    @classmethod
    def convert(cls, *args, **kwargs):
        return Report.convert(*args, **kwargs)


class PrintShippingLabel(metaclass=PoolMeta):
    __name__ = 'stock.shipment.print_shipping_label'

    @classmethod
    def __setup__(cls):
        super().__setup__()
        cls.print_.action_id = 'jurassic.stock_package_shipping_label_report'


class PackageShippingLabelGrouped(Report):
    __name__ = 'stock.package.shipping_label.grouped'

    @classmethod
    def header_key(cls, record):
        Move = namedtuple('Move', ['product', 'quantity', 'unit'])
        return (
            ('moves', sorted([Move(m.product, m.quantity, m.unit)
                        for m in record.moves])),
            )


class ShipmentPackageShippingLabel(Wizard):
    "Shipment Package Shipping Label"
    __name__ = 'stock.shipment.shipping_label'

    start = StateAction('jurassic.stock_package_shipping_label_grouped_report')

    def do_start(self, action):
        pool = Pool()
        PackageShippingLabelGrouped = pool.get(
            'stock.package.shipping_label.grouped', type='report')
        packages = [p for s in self.records for p in s.packages]
        packages = sorted(packages, key=PackageShippingLabelGrouped.header_key)
        ids = [p.id for p in packages]
        return action, {
            'ids': ids,
            }


class ProductsByLocations(metaclass=PoolMeta):
    __name__ = 'stock.products_by_locations'

    made_in = fields.Many2One('country.country', "Made In")


class ProductPosition(ModelSQL, ModelView):
    "Product Position"
    __name__ = 'stock.product.position'

    warehouse = fields.Many2One(
        'stock.location', "Warehouse",
        domain=[
            ('type', '=', 'warehouse'),
            ],
        required=True)
    product = fields.Many2One(
        'product.product', "Product",
        domain=[
            ('type', '=', 'goods'),
            ],
        required=True)
    x = fields.Integer(
        "X",
        domain=[If(Eval('excluded', False), [
                    ('x', '=', None),
                    ], [
                    ('x', '>=', 0),
                    ('x', '<=', Eval('max_x', 0)),
                    ])],
        states={
            'required': ~Eval('excluded', False),
            'invisible': Eval('excluded', False),
            },
        depends=['max_x', 'excluded'])
    y = fields.Integer(
        "Y",
        domain=[If(Eval('excluded', False), [
                    ('y', '=', None),
                    ], [
                    ('y', '>=', 0),
                    ('y', '<=', Eval('max_y', 0)),
                    ])],
        states={
            'required': ~Eval('excluded', False),
            'invisible': Eval('excluded', False),
            },
        depends=['max_y', 'excluded'])
    y_string = fields.Function(fields.Char("Y"), 'on_change_with_y_string')
    static = fields.Boolean(
        "Static",
        domain=[If(Eval('excluded', False),
                ('static', '=', False),
                ())],
        states={
            'invisible': Eval('excluded', False),
            },
        depends=['excluded'])
    excluded = fields.Boolean("Excluded")
    position = fields.Char(
        "Position",
        domain=[If(~Eval('excluded', False),
                ['OR',
                    ('position', '=', ''),
                    ('position', '=', None),
                    ],
                [])],
        states={
            'invisible': ~Eval('excluded'),
            },
        depends=['excluded'])

    max_x = fields.Function(fields.Integer("Max X"), 'on_change_with_max_x')
    max_y = fields.Function(fields.Integer("Max Y"), 'on_change_with_max_y')

    @classmethod
    def __setup__(cls):
        super().__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('warehouse_product_unique',
                Unique(t, t.warehouse, t.product),
                'jurassic.msg_stock_product_position_'
                'warehouse_product_unique'),
            ('warehouse_x_y_unique',
                Unique(t, t.warehouse, t.x, t.y),
                'jurassic.msg_stock_product_position_'
                'warehouse_position_unique'),
            ('warehouse_position_unique',
                Unique(t, t.warehouse, t.position),
                'jurassic.msg_stock_product_position_'
                'warehouse_position_unique'),
            ]
        cls._sql_indexes.add(
            Index(
                t,
                (t.warehouse, Index.Equality()),
                (t.product, Index.Equality())))

    @fields.depends('y')
    def on_change_with_y_string(self, name=None):
        if self.y is not None:
            return int2base(self.y, string.ascii_uppercase)

    @fields.depends('warehouse')
    def on_change_with_max_x(self, name=None):
        pool = Pool()
        Configuration = pool.get('stock.configuration')
        config = Configuration(1)
        return config.get_multivalue(
            'max_x', warehouse=self.warehouse.id if self.warehouse else None)

    @fields.depends('warehouse')
    def on_change_with_max_y(self, name=None):
        pool = Pool()
        Configuration = pool.get('stock.configuration')
        config = Configuration(1)
        return config.get_multivalue(
            'max_y', warehouse=self.warehouse.id if self.warehouse else None)

    def get_rec_name(self, name):
        if self.x is not None and self.y is not None:
            return '%d-%s' % (self.x, self.y_string)
        elif self.position:
            return self.position
        else:
            return ''


class ProductPositionCompute(Wizard):
    "Compute Product Position"
    __name__ = 'stock.product.position.compute'

    start = StateView(
        'stock.product.position.compute.start',
        'jurassic.stock_product_position_compute_start_view_form', [
            Button("Cancel", 'end'),
            Button("Compute", 'compute', default=True),
            ])
    compute = StateTransition()

    def transition_compute(self):
        pool = Pool()
        Configuration = pool.get('stock.configuration')
        ProductPosition = pool.get('stock.product.position')
        Move = pool.get('stock.move')
        move = Move.__table__()
        cursor = Transaction().connection.cursor()
        config = Configuration(1)

        positions = ProductPosition.search([
                ('warehouse', '=', self.start.warehouse.id),
                ('static', '=', False),
                ('excluded', '=', False),
                ])
        ProductPosition.delete(positions)
        statics = ProductPosition.search([
                ('warehouse', '=', self.start.warehouse.id),
                ('static', '=', True),
                ])
        static_products = {s.product.id for s in statics}
        static_positions = {(s.x, s.y) for s in statics}
        excluded = ProductPosition.search([
                ('warehouse', '=', self.start.warehouse.id),
                ('excluded', '=', True),
                ])
        excluded_products = {e.product.id for e in excluded}

        output_location = self.start.warehouse.output_location

        query = move.select(
            move.product,
            where=(move.from_location == output_location.id)
            & (move.planned_date >= self.start.from_)
            & (move.planned_date <= self.start.to)
            & (move.quantity > 0),
            group_by=[move.product],
            order_by=Count(Literal('*')).desc)
        cursor.execute(*query)

        def next_position(x, y):
            y += 1
            if y > max_y:
                y = 0
                x += 1
            return x, y

        def real_x(x):
            return (max_x // 2) + (x // 2) * (-1) ** x

        positions = []
        max_x, max_y = config.max_x or 0, config.max_y or 0
        x, y = 1, 0
        for product_id, in cursor:
            if (product_id in static_products
                    or product_id in excluded_products):
                continue
            while (real_x(x), y) in static_positions:
                x, y = next_position(x, y)
            positions.append(ProductPosition(
                    warehouse=self.start.warehouse,
                    product=product_id,
                    x=real_x(x),
                    y=y))
            x, y = next_position(x, y)
        ProductPosition.save(positions)
        return 'end'


class ProductPositionComputeStart(ModelView):
    "Compute Product Position"
    __name__ = 'stock.product.position.compute.start'

    warehouse = fields.Many2One(
        'stock.location', "Warehouse",
        domain=[
            ('type', '=', 'warehouse'),
            ],
        required=True)
    from_ = fields.Date("From", required=True)
    to = fields.Date("To", required=True)

    @classmethod
    def default_warehouse(cls):
        Location = Pool().get('stock.location')
        locations = Location.search(cls.warehouse.domain)
        if len(locations) == 1:
            return locations[0].id

    @classmethod
    def default_from_(cls):
        return Pool().get('ir.date').today()

    default_to = default_from_


class ShipmentCostDefaultWeightMixin:
    __slots__ = ()

    @property
    def allocation_shipment_cost_method(self):
        method = super().allocation_shipment_cost_method
        if not getattr(self, 'carrier', None):
            method = 'weight'
        return method
